<?php /****************************************************************************************************************
* This Source Code Form is subject to the terms of the Mozilla Public
* License, v. 2.0. If a copy of the MPL was not distributed with this
* file, You can obtain one at http://mozilla.org/MPL/2.0/.
**********************************************************************************************************************/

// == | Setup and Global Constants | ======================================================================

// Check for kRootPath
if (!defined('kRootPath')) {
  die('BinOC Metropolis Utilities: You MUST define kRootPath.');
}

// Do not allow this to be included more than once...
if (defined('kUtilities')) {
  die('BinOC Metropolis Utilities: You may not include this file more than once.');
}

// Define that this is a thing which can double as a version check.
define('kUtilities', '2.1.0a1');

// --------------------------------------------------------------------------------------------------------------------

// These constants must be defined because they might be used before nsRuntime is init'd
if (!defined('kAppVendor'))             { define('kAppVendor', 'Binary Outcast'); }
if (!defined('kAppName'))               { define('kAppName', 'Metropolis-based Software'); }
if (!defined('kAppVersion'))            { define('kAppVersion', kUtilities); }
if (!defined('kAppRepository'))         { define('kAppRepository', '#'); }
if (!defined('kUtilsGlobalWrappers'))   { define('kUtilsGlobalWrappers', true); }
if (!defined('kBootstrap'))             { define('kBootstrap', true); }
if (!defined('kDefaultSkinPath'))       { define('kDefaultSkinPath', './base/skin'); }

// --------------------------------------------------------------------------------------------------------------------

// We like CLI
define('SAPI_IS_CLI', php_sapi_name() == "cli");
define('CLI_NO_LOGO', in_array('--nologo', $GLOBALS['argv'] ?? []));

// Enable PHP Error Reporting
error_reporting(E_ALL);
ini_set('display_errors', true);
ini_set('display_startup_errors', true);
const E_EXCEPTION = 65536;

// Debug flag (CLI always triggers debug mode)
define('kDebugMode', $_GET['debug'] ?? SAPI_IS_CLI);

// --------------------------------------------------------------------------------------------------------------------

// Define basic symbol constants
const kNewLine              = "\n";
const kCR                   = "\r";
const kLF                   = "\n";
const kTab                  = "\t";
const kSpace                = " ";
const kQuote                = '"';
const kApostrophe           = "'";
const kLessThan             = "<";
const kGreaterThan          = ">";
const kAmpersand            = "&";
const kSlash                = "/";
const kBackSlash            = "\\";
const kEqual                = "=";
const kDash                 = "-";
const kPlus                 = "+";
const kExclamation          = "!";
const kSemicolon            = ";";
const kHashsign             = "#";
const kAsterisk             = "*";
const kUnderbar             = "_";
const kComma                = ",";
const kLeftParen            = "(";
const kRightParen           = ")";
const kLeftBrace            = "{";
const kRightBrace           = "}";
const kQuestionMark         = "?";
const kLeftSquareBracket    = "[";
const kRightSquareBracket   = "]";
const kNullCh               = "\0";

const kForwardSlash         = kSlash;
const kMinus                = kDash;

const kDot                  = ".";
const kPipe                 = "|";
const kDollar               = "\$";
const kColon                = ":";

const kScopeOperator        = kColon . kColon;
const kResolutionOperator   = kDash . kGreaterThan;
const kDotDot               = kDot . kDot;
const kSpaceDashSpace       = kSpace . kDash . kSpace;
const kColonSlashSlash      = kColon . kSlash . kSlash;

const kEmptyString          = "";
const kEmptyArray           = [];

const kNegOne               = -1;
const kZero                 = 0;
const kPosOne               = 1;

// --------------------------------------------------------------------------------------------------------------------

const kJsonPrettyEncode     = nsUtils::JSON_ENCODE_FLAGS['display'];

// --------------------------------------------------------------------------------------------------------------------

const kDefaultSkinName      = 'default';
const kDefaultMenu          = ['/' => 'Front Page (Home)'];
const kSpecialComponent     = 'special';
const kSpecialComponentName = 'Special Component';

const PALEMOON_GUID         = '{8de7fcbb-c55c-4fbe-bfc5-fc555c87dbc4}';

// --------------------------------------------------------------------------------------------------------------------

# Include vc and adbar-dot
require_once(dirname(__FILE__) . "/nsUtilsDotNotation.php");
require_once(dirname(__FILE__) . "/nsUtilsDotNotation.php");

// ====================================================================================================================

// == | Static Application Class | ====================================================================================

class nsRuntime {
  private static $sInitialized = false;
  private static $sRuntimeConfiguration = kEmptyArray;
  public static $sModule = kEmptyArray;
  const VIRTUAL_NODES         = ['constant', 'superglobal'];
  
  /********************************************************************************************************************
  * Application Init
  *********************************************************************************************************************/
  public static function init() {
    // Gaurd against being init'd again
    if (nsRuntime::$sInitialized) {
      return nsRuntime::$sInitialized;
    }

    // ------------------------------------------------------------------------------------------------------------------

    // Explode the path
    $path = nsUtils::ExplodePath(nsRuntime::SuperGlobal('get', 'path', kSlash));

    // Handle localhost
    if (defined('kLocalDomain') && SAPI_IS_CLI) {
      $localDomain = kLocalDomain;
    }

    // Set a configuration array
    $config = array(
      'app' => array(
        'component'   => nsRuntime::SuperGlobal('get', 'component', 'site'),
        'path'        => $path,
        'depth'       => count($path ?? kEmptyArray),
        'debug'       => kDebugMode,
        'offline'     => file_exists(nsUtils::BuildPath(kRootPath, '.offline')),
      ),
      'network' => array(
        'scheme'      => nsRuntime::SuperGlobal('server', 'SCHEME') ?? (nsRuntime::SuperGlobal('server', 'HTTPS') ? 'https' : 'http'),
        'serverName'  => nsRuntime::SuperGlobal('server', 'SEVER_NAME', ($localDomain ?? 'localhost')),
        'requestUri'  => nsRuntime::SuperGlobal('server', 'REQUEST_URI', kSlash),
        'requestPath' => nsRuntime::SuperGlobal('get', 'path', kSlash),
        'remoteAddr'  => nsRuntime::SuperGlobal('server', 'HTTP_X_FORWARDED_FOR', nsRuntime::SuperGlobal('server', 'REMOTE_ADDR', '127.0.0.1')),
        'userAgent'   => nsRuntime::SuperGlobal('server', 'HTTP_USER_AGENT', 'php' . kDash . PHP_SAPI . kSlash . PHP_VERSION),
        'baseDomain'  => nsConsole::GetDomain(nsRuntime::SuperGlobal('server', 'SERVER_NAME', ($localDomain ?? 'localhost'))),
        'subDomain'   => nsConsole::GetDomain(nsRuntime::SuperGlobal('server', 'SERVER_NAME', ($localDomain ?? 'localhost')), true),
      ),
      'console' => array(
        'output' => array(
          'contentType' => 'text/plain',
          'responseCode'  => 200,
          'responseBody'  => kEmptyString,
          'httpHeaders' => kEmptyArray,
        ),
        'ui' => array(
          'mainmenu'      => kDefaultMenu,
          'commandbar'    => kDefaultMenu,
          'statustext'    => 'Done',
        ),
      ),
    );

    // Handle debug mode
    if (defined('kDebugDomain') && !SAPI_IS_CLI) {
      $config['app']['debug'] = (nsRuntime::SuperGlobal('server', 'SERVER_NAME', 'localhost') != constant('kDebugDomain') ?? kEmptyString) ?
                                file_exists(nsUtils::BuildPath(kRootPath, '.debugMode')) :
                                !kDebugMode;
    }

    // ------------------------------------------------------------------------------------------------------------------

    // Set runtime configuration to our inital configuration
    nsRuntime::$sRuntimeConfiguration = new \Adbar\Dot($config);

    // ------------------------------------------------------------------------------------------------------------------
    
    // Set error handlers
    nsConsole::EnableErrorHandler();

    // ------------------------------------------------------------------------------------------------------------------

    // These are global wrapping functions.
    if (kUtilsGlobalWrappers) {
      // Application
      function gError                       (...$args) { return nsRuntime::Error(...$args); }
      function gNotFound                    (...$args) { return nsRuntime::NotFound(...$args); }
      function gReadFile                    (...$args) { return nsRuntime::ReadFile(...$args); }
      function gGetConfig                   (...$args) { return nsRuntime::GetConfig(...$args); }
      function gSetConfig                   (...$args) { return nsRuntime::SetConfig(...$args); }
      function gEnsureValue                 (...$args) { return nsRuntime::EnsureValue(...$args); }
      function gRegisterIncludes            (...$args) { return nsRuntime::RegisterIncludes(...$args); }
      function gImportModules               (...$args) { return nsRuntime::ImportModules(...$args); }
      function gLoadComponent               (...$args) { return nsRuntime::LoadComponent(...$args); }
      function gRegistry                    (...$args) { return nsRuntime::GetConfig(...$args); }
      function gRegSet                      (...$args) { return nsRuntime::SetConfig(...$args); }

      // Application Utilities
      function gSubst                       (...$args) { return nsUtils::Subst(...$args); }
      function gSubstEx                     (...$args) { return nsUtils::SubstEx(...$args); }
      function gContains                    (...$args) { return nsUtils::Contains(...$args); }
      function gExplodeStr                  (...$args) { return nsUtils::ExplodeStr(...$args); }
      function gBuildPath                   (...$args) { return nsUtils::BuildPath(...$args); }
      function gStripStr                    (...$args) { return nsUtils::StripStr(...$args); }
      function gHexString                   (...$args) { return nsUtils::HexString(...$args); }
      function gPasswordHash                (...$args) { return nsUtils::PasswordHash(...$args); }
      function gPasswordVerify              (...$args) { return nsUtils::PasswordVerify(...$args); }
      function gGlobalIdentifer             (...$args) { return nsUtils::GlobalIdentifer(...$args); }

      // Console
      function gHeader                      (...$args) { return nsConsole::Header(...$args); }
      function gContentType                 (...$args) { return nsConsole::ContentType(...$args); }
      function gSendHeaders                 (...$args) { return nsConsole::SendHeaders(...$args); }
      function gRedirect                    (...$args) { return nsConsole::Redirect(...$args); }
      function gOutput                      (...$args) { return nsConsole::Output(...$args); }
      function gContent                     (...$args) { return nsConsole::Content(...$args); }
    }

    // ------------------------------------------------------------------------------------------------------------------

    nsRuntime::$sInitialized = true;

    // This will bootstrap the Metropolis-based application loading. You may set kBoostrap to false in the entry point to
    // completely bypass this execution and return to the entry point.
    //
    // Bootstrap has THREE modes of operation:
    // App is Special "Component"     - Simply forces the application to always load the Special "Component".
    // App is ./base/src/nsWebApp.php     - In this mode we will load the application from this file and if execution isn't
    //                                  otherwise terminated it will terminate here with a 404/PC LOAD LETTER.
    // App is Entry point             - Return to the entry point but allow for the possibility of switching to one of
    //                                  the other modes without adding or removing kBootstrap
    //
    // NOTE: Future behavior will have kBootstrap set to false by default.. like in a few weeks if not sooner.
    if (kBootstrap) {
      // The application effectively IS the special component
      if (nsRuntime::GetConfig('constant.appIsSpecialComponent')) {
        nsRuntime::LoadComponent(kSpecialComponent);
      }

      // App is ./base/src/nsWebApp.php
      if (file_exists(nsUtils::BuildPath(kRootPath, 'base', 'src', 'nsInternetApplication.php'))) {
        require_once(nsUtils::BuildPath(kRootPath, 'base', 'src', 'nsInternetApplication.php'));

        if (nsRuntime::GetConfig('app.path.0') == kSpecialComponent) {
          nsRuntime::SetConfig('app.component', kSpecialComponent);
        }

        nsRuntime::LoadComponent(nsRuntime::GetConfig('app.component'));
        nsRuntime::NotFound('PC LOAD LETTER');
      }
    }
  }

  /********************************************************************************************************************
  * Set app config key
  *********************************************************************************************************************/
  public static function GetConfig(string $aDottedKey, mixed $aFallback = null) {
    $rv = null;

    if (nsUtils::Contains($aDottedKey, nsRuntime::VIRTUAL_NODES, nsUtils::STARTS_WITH)) {
      $keyNodes = nsUtils::ExplodeStr(kDot, $aDottedKey) ?? kEmptyArray;
      $firstNode = $keyNodes[array_key_first($keyNodes)] ?? kEmptyString;
      $dot = new \Adbar\Dot();
      switch($firstNode) {
        case 'constant':
          if (count($keyNodes) < 2) { return $aFallback; }

          $ucConst = strtoupper($keyNodes[1]);
          $prefixConst = 'k' . ucfirst($keyNodes[1]);

          switch (true) {
            case defined($prefixConst):
              $rv = constant($prefixConst);
              break;
            case defined($ucConst):
              $rv = constant($ucConst);
              break;
            case defined($keyNodes[1]):
              $rv = constant($keyNodes[1]);
              break;
            default:
              return null;
          }

          if (!(is_array($rv) || $rv instanceof ArrayAccess)) { return $rv ?? $aFallback; }
          unset($keyNodes[0], $keyNodes[1]);
          $rv = $dot->setArray($rv)->get(implode(kDot, $keyNodes), $aFallback);
          break;
        case 'superglobal':
          if (count($keyNodes) < 3) { return $aFallback; }
          $rv = nsRuntime::SuperGlobal($keyNodes[1], $keyNodes[2]);
          if (!(is_array($rv) || $rv instanceof ArrayAccess)) { return $rv ?? $aFallback; }
          unset($keyNodes[0], $keyNodes[1]);
          $rv = $dot->setArray($rv)->get(implode(kDot, $keyNodes), $aFallback);
          break;
        default:
          return null;
      }
    }
    else { $rv = nsRuntime::$sRuntimeConfiguration->get($aDottedKey, $aFallback); }

    return $rv; 
  }

  /********************************************************************************************************************
  * Set app config key
  *********************************************************************************************************************/
  public static function SetConfig(string $aDottedKey, mixed $aNewValue) {
    $keyNodes = nsUtils::ExplodeStr(kDot, $aDottedKey) ?? kEmptyArray;
    $firstNode = $keyNodes[array_key_first($keyNodes)] ?? kEmptyString;
    $lastNode = $keyNodes[array_key_last($keyNodes)] ?? kEmptyString;

    if (nsUtils::Contains(nsRuntime::VIRTUAL_NODES, $firstNode, nsUtils::IN_ARRAY)) {
      nsRuntime::Error('Setting values on virtual nodes is not supported.');
    }

    if (nsUtils::Contains($lastNode, '[]', nsUtils::ENDS_WITH)) {
      $aDottedKey = substr($aDottedKey, 0, -2);
      
      $oldValue = nsRuntime::$sRuntimeConfiguration->get($aDottedKey, kEmptyArray);

      if ($oldValue && !is_array($oldValue)) {
        $oldValue = [$oldValue];
      }

      $oldValue[] = $aNewValue;
      $aNewValue = $oldValue;
    }
    else if (is_numeric($lastNode)) {
      array_pop($keyNodes);
      $truncatedDottedKey = implode(kDot, $keyNodes);
      $oldValue = nsRuntime::$sRuntimeConfiguration->get(implode(kDot, $keyNodes));

      if (!$oldValue || !array_is_list($oldValue)) {
        nsRuntime::Error('Using indexed keys can only be done if the value is already an indexed list.');
      }

      $oldValue[(int)$lastNode] = $aNewValue;
      $aNewValue = $oldValue;
      $aDottedKey = $truncatedDottedKey;
    }

    return nsRuntime::$sRuntimeConfiguration->set($aDottedKey, $aNewValue);
  }

  /********************************************************************************************************************
  * Loads a component.
  *********************************************************************************************************************/
  public static function LoadComponent(string $aComponent) {
    if ($aComponent == kSpecialComponent) {
      nsRuntime::SpecialComponent();
    }

    $componentPath = nsRuntime::GetConfig('constant.components' . kDot . $aComponent);

    if (!$componentPath) {
      nsRuntime::NotFound('Unknown component.');
    }

    if (!file_exists($componentPath)) {
      nsRuntime::NotFound('Failed to load the' . kSpace . $aComponent . kSpace .'component.');
    }

    nsRuntime::SetConfig('app.componentPath', nsUtils::BuildPath(kRootPath, 'components', $aComponent));
    require_once($componentPath);
  }

  /********************************************************************************************************************
  * Read a file
  *********************************************************************************************************************/
  public static function ReadFile(string $aFile) {
    $rv = @file_get_contents($aFile);
    return nsRuntime::EnsureValue($rv);
  }

  /********************************************************************************************************************
  * General Error Function
  *
  * @param $aMessage   Error message
  ********************************************************************************************************************/
  public static function Error(mixed $aMessage = null) {
    if (!$aMessage) {
      $aMessage = 'No further details were provided.';
    }

    if (!is_string($aMessage)) {
      nsConsole::Output($aMessage);
    }

    nsConsole::Report(['code' => E_ALL, 'message' => $aMessage,
                         'file' => null, 'line' => null,
                         'trace' => debug_backtrace(2)]);
  }

  /********************************************************************************************************************
  * Sends 404 or prints error message if debug mode
  ********************************************************************************************************************/
  public static function NotFound(?string $aMessage = null) {
    if (!$aMessage) {
      $aMessage = 'HTTP/1.1 404 Not Found';
    }

    if (nsRuntime::Debug()) {
      nsConsole::Report(['code' => E_ALL, 'message' => $aMessage,
                           'file' => null, 'line' => null,
                           'trace' => debug_backtrace(2)]);
    }
    nsConsole::Header(404);
  }

  /********************************************************************************************************************
  * Check if a value should be null
  ********************************************************************************************************************/
  public static function EnsureValue($aValue, $aFallback = null) {
    return (empty($aValue) || $aValue === 'none') ? $aFallback : $aValue;
  }

  /**********************************************************************************************************************
  * Special Component!
  ***********************************************************************************************************************/
  public static function SpecialComponent() {
    $spCurrentPath = nsRuntime::GetConfig('app.path');
    $spPathCount = nsRuntime::GetConfig('app.depth');

    if ($spCurrentPath[0] != kSpecialComponent) {
      nsConsole::Redirect(kSlash . kSpecialComponent . kSlash);
    }

    nsRuntime::SetConfig('app.component', kSpecialComponent);

    if (nsRuntime::GetConfig('constant.disableSpecialComponent')) {
      nsRuntime::NotFound('The special component has been disabled.');
    }

    nsRuntime::SetConfig('console.ui.sectionName', kSpecialComponentName);

    // The Special Component never has more than one level below it
    // We still have to determine the root of the component though...
    if ($spPathCount == 1) {
      // URL /special/
      $spSpecialFunction = 'root';
    }
    else {
      // URL /special/xxx/
      if ($spPathCount > 2) {
        nsRuntime::NotFound('The special component only has one path level.');
      }
      $spSpecialFunction = $spCurrentPath[1];
    }

    $spCommandBar = array(
      '/special/'                 => kSpecialComponentName,
      '/special/test/'            => 'Test Cases',
      '/special/guid/'            => 'GUID',
      '/special/hex/'             => 'Hex String',
    );

    nsRuntime::SetConfig('console.ui.commandbar', nsRuntime::GetConfig('constant.components.site') ?
                                               array_merge(kDefaultMenu, $spCommandBar) :
                                               $spCommandBar);

    unset($spCurrentPath, $spPathCount, $spCommandBar);

    switch ($spSpecialFunction) {
      case 'root':
        $spContent = '<h2>Welcome</h2>' .
                     '<p>Please select a special function from the command bar above.';
        nsConsole::Content($spContent, ['title' => 'Overview']);
        break;
      case 'test':
        if (!nsRuntime::Debug()) {
          nsRuntime::NotFound('This special function is not available when not in debug mode.');
        }
        $spCase = nsRuntime::GetConfig('superglobal.get.case');
        $spTestsPath = nsUtils::BuildPath(kRootPath, 'base', 'tests');
        $spGlobTests = glob(nsUtils::BuildPath($spTestsPath, kAsterisk . nsUtils::FILE_EXT['php']));
        $spTests = kEmptyArray;

        foreach ($spGlobTests as $_value) {
          $spTests[] = nsUtils::Subst($_value, [nsUtils::FILE_EXT['php'] => kEmptyString, $spTestsPath . kSlash => kEmptyString]);
        }

        if ($spCase) {
          if (!nsUtils::Contains($spCase, $spTests)) {
            nsRuntime::Error('Unknown test case.');
          }

          nsRuntime::SetConfig('special.testCase', $spCase);
          require_once(nsUtils::BuildPath($spTestsPath, $spCase . nsUtils::FILE_EXT['php']));
          headers_sent() ? exit() : nsRuntime::Error('The operation completed successfully.');
        }

        $spContent = kEmptyString;

        foreach ($spTests as $_value) {
          $spContent .= '<li><a href="/special/test/?case=' . $_value . '">' . $_value . '</a></li>';
        }

        $spContent = ($spContent == kEmptyString) ?
                     '<p>There are no test cases.</p>' :
                     '<h2>Please select a test case&hellip;</h2><ul>' . $spContent . '</ul>' . str_repeat('<br />', 3);

        nsConsole::Content($spContent, ['title' => 'Test Cases']);
        break;
      case 'guid':
        nsConsole::Content(nsUtils::GlobalIdentifer(nsRuntime::GetConfig('superglobal.get.vendor'), true),
                 ['title' => 'Globally Unique Identifier (In XPIDL Notation)', 'textbox' => true]);
        break;
      case 'hex':
        nsConsole::Content(nsUtils::HexString(nsRuntime::GetConfig('superglobal.get.length', 40)),
                 ['title' => 'Pseudo-Random Hex String', 'textbox' => true]);
        break;
      case 'system':
        if (!nsRuntime::Debug()) {
          nsRuntime::NotFound('This special function is not available when not in debug mode.');
        }
        nsRuntime::Header('html', true);
        phpinfo(INFO_GENERAL | INFO_CONFIGURATION | INFO_ENVIRONMENT | INFO_VARIABLES);
        break;
      default:
        nsRuntime::NotFound('There is no matching function in the special component.');
    }

    // We're done here
    exit();
  }
  
  /********************************************************************************************************************
  * Access Super Globals
  ********************************************************************************************************************/
  public static function SuperGlobal($aNode, $aKey, $aDefault = null) {
    $rv = null;

    // Turn the variable type into all caps prefixed with an underscore
    $aNode = kUnderbar . strtoupper($aNode);

    // This handles the superglobals
    switch($aNode) {
      case '_CHECK':
        $rv = nsRuntime::EnsureValue($aKey);
        break;
      case '_GET':
        if (SAPI_IS_CLI && $GLOBALS['argc'] > 1) {
          $args = kEmptyArray;

          foreach (array_slice($GLOBALS['argv'], 1) as $_value) {
            $arg = @explode('=', $_value);

            if (count($arg) < 2) {
              continue;
            }

            $attr = str_replace('--', kEmptyString, $arg[0]);
            $val = nsRuntime::EnsureValue(str_replace('"', kEmptyString, $arg[1]));

            if (!$attr && !$val) {
              continue;
            }

            $args[$attr] = $val;
          }

          $rv = $args[$aKey] ?? $aDefault;
          break;
        }
      case '_SERVER':
      case '_ENV':
      case '_FILES':
      case '_POST':
      case '_COOKIE':
      case '_SESSION':
        $rv = $GLOBALS[$aNode][$aKey] ?? $aDefault;
        break;
      default:
        // We don't know WHAT was requested but it is obviously wrong...
        nsRuntime::Error('Unknown system node.');
    }
    
    // We always pass $_GET values through a general regular expression
    // This allows only a-z A-Z 0-9 - / { } @ % whitespace and ,
    if ($rv && $aNode == "_GET") {
      $rv = preg_replace(nsUtils::REGEX_PATTERNS['query'], kEmptyString, $rv);
    }

    // Files need special handling.. In principle we hard fail if it is anything other than
    // OK or NO FILE
    if ($rv && $aNode == "_FILES") {
      if (!in_array($rv['error'], [UPLOAD_ERR_OK, UPLOAD_ERR_NO_FILE])) {
        nsRuntime::Error('Upload of ' . $aKey . ' failed with error code: ' . $rv['error']);
      }

      // No file is handled as merely being null
      if ($rv['error'] == UPLOAD_ERR_NO_FILE) {
        return null;
      }

      // Cursory check the actual mime-type and replace whatever the web client sent
      $rv['type'] = mime_content_type($rv['tmp_name']);
    }

    return $rv;
  }

  /********************************************************************************************************************
  * Registers Files to be included such as components and modules
  *********************************************************************************************************************/
  public static function RegisterIncludes($aConst, $aIncludes) {
    $aConst = strtoupper($aConst);

    if (defined($aConst)) {
      nsRuntime::Error($aConst . kSpace . 'files are already registered and may not be updated.');
    }

    $includes = kEmptyArray;

    foreach($aIncludes as $_key => $_value) { 
      switch ($aConst) {
        case 'COMPONENTS':
          $includes[$_value] = nsUtils::BuildPath(kRootPath, 'components', $_value, 'src', 'nsComponent' . nsUtils::FILE_EXT['php']);
          break;
        case 'MODULES':
          $includes[$_value] = nsUtils::BuildPath(kRootPath, 'modules', 'ns' . ucfirst($_value) . nsUtils::FILE_EXT['php']);
          break;
        case 'LIBRARIES':
          if (str_contains($_value, kDot . kDot)) {
            return;
          }

          $includes[$_key] = nsUtils::BuildPath(kRootPath, 'third_party', $_value);
          break;
        default:
          gError('Unknown include type');
      }
    }

    define($aConst, $includes);
  }

  /**********************************************************************************************************************
  * Includes a module
  **********************************************************************************************************************/
  public static function ImportModules(...$aModules) {
    if (!defined('MODULES')) {
      funcError('MODULES is not defined');
    }

    foreach ($aModules as $_value) {
      if (!array_key_exists($_value, MODULES)) {
        gError('Unable to import unknown module ' . $_value);
      }

      $className = 'ns' . ucfirst($_value);
      $moduleName = $_value;
   
      if (array_key_exists($moduleName, nsRuntime::$sModule)) {
        gError('Module ' . $_value . ' has already been imported');
      }

      require(MODULES[$_value]);
      nsRuntime::$sModule[$moduleName] = new $className();
    }
  }

  /********************************************************************************************************************
  * Get the registry property and return it
  ********************************************************************************************************************/
  public static function Component(?string $aCompareComponent = null) {
    $rv = (nsRuntime::$sRuntimeConfiguration) ? nsRuntime::GetConfig('app.component') : nsRuntime::SuperGlobal('get', 'component', 'site');

    if ($aCompareComponent) {
      $rv = ($rv === $aCompareComponent);
    }

    return $rv;
  }

  /********************************************************************************************************************
  * Get the debyg property and return it
  ********************************************************************************************************************/
  public static function Debug() {
    return (nsRuntime::$sRuntimeConfiguration) ? nsRuntime::GetConfig('app.debug') : kDebugMode;
  }

  /********************************************************************************************************************
  * Get the registry and return it
  ********************************************************************************************************************/
  public static function GetStore() {
    return nsRuntime::$sRuntimeConfiguration->get();
  }
}

// ====================================================================================================================

// == | Application Utilities | =======================================================================================

class nsUtils {
  const CONTAINS = kZero;
  const STARTS_WITH = kPosOne;
  const ENDS_WITH = kNegOne;
  const IN_ARRAY = kZero;
  const KEY_EXISTS = kPosOne;

  // --------------------------------------------------------------------------------------------------------------------

  const PASSWORD_CLEARTEXT    = "clrtxt";
  const PASSWORD_HTACCESS     = "apr1";
  const BASE64_ALPHABET       = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
  const APRMD5_ALPHABET       = './0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz';

  // --------------------------------------------------------------------------------------------------------------------

  const FILE_WRITE_FLAGS      = "w+";
  const FILE_EXT             = array(
    'php'                     => kDot . 'php',
    'ini'                     => kDot . 'ini',
    'html'                    => kDot . 'html',
    'xhtml'                   => kDot . 'xhtml',
    'xml'                     => kDot . 'xml',
    'rdf'                     => kDot . 'rdf',
    'json'                    => kDot . 'json',
    'content'                 => kDot . 'content',
    'tpl'                     => kDot . 'tpl',
    'xpinstall'               => kDot . 'xpi',
    'jar'                     => kDot . 'jar',
    'winstaller'              => kDot . 'installer' . kDot . 'exe',
    'winportable'             => kDot . 'portable' . kDot . 'exe',
    'mar'                     => kDot . 'mar',
    'mar-bz2'                 => kDot . 'complete' . kDot . 'mar',
    '7z'                      => kDot . '7z',
    'zip'                     => kDot . 'zip',
    'tz'                      => kDot . 'tz',
    't7z'                     => kDot . 't7z',
    'tgz'                     => kDot . 'tgz',
    'tbz'                     => kDot . 'tbz',
    'txz'                     => kDot . 'txz',
    'tar-zip'                 => kDot . 'tar' . kDot . 'zip',
    'tar-7z'                  => kDot . 'tar' . kDot . '7z',
    'tar-gz'                  => kDot . 'tar' . kDot . 'gz',
    'tar-bz2'                 => kDot . 'tar' . kDot . 'bz2',
    'tar-xz'                  => kDot . 'tar' . kDot . 'xz', 
  );

  // ----------------------------------------------------------------------------------------------------------------

  const TAG_XML               = '<?xml version="1.0" encoding="utf-8" ?>';

  // ----------------------------------------------------------------------------------------------------------------

  const JSON_ENCODE_FLAGS            = array(
    'display'                 => JSON_PRETTY_PRINT | JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE,
    'storage'                 => JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE,
  );

  // ----------------------------------------------------------------------------------------------------------------

  const REGEX_PATTERNS        = array(
    'query'                   => "/[^-a-zA-Z0-9_\-\/\{\}\@\.\%\s\,]/",
    'yaml'                    => "/\A---(.|\n)*?---/",
    'guid'                    => "/^\{[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\}$/i",
    'host'                    => "/[a-z0-9-\._]+\@[a-z0-9-\._]+/i",
  );

  // ------------------------------------------------------------------------------------------------------------------

  /**********************************************************************************************************************
  * Do "Filter Subtsitution" of a string
  ***********************************************************************************************************************/
  public static function Subst(string $aString, array $aSubsts, bool $aRegEx = false) {
    $rv = $aString;
    $replaceFunction = $aRegEx ? 'preg_replace' : 'str_replace';

    foreach ($aSubsts as $_key => $_value) {
      $rv = call_user_func($replaceFunction, ($aRegEx ? kSlash . $_key . kSlash . 'iU' : $_key), $_value, $rv);
    }

    return !$rv ? nsRuntime::Error('Something has gone wrong...') : $rv;
  } 
    
  /**********************************************************************************************************************
  * Do advanced "Filter Substitution" of a string
  ***********************************************************************************************************************/
  public static function SubstEx(string $aString, ...$aSubsts) {
    $keyPrefix = kLeftBrace;
    $keySuffix = kRightBrace;
    $substs = $aSubsts;

    // Accept an array as the second argument for named placeholder substitution further arguments or
    // changing the pre/post prefix.
    if (is_array($aSubsts[0])) {
      $substs = $aSubsts[0];
      $keyPrefix = $aSubsts[1] ?? kLeftBrace;
      $keySuffix = $aSubsts[2] ?? kRightBrace;
    }

    $keyCallback = function ($aKey) use ($keyPrefix, $keySuffix) { return $keyPrefix . $aKey . $keySuffix; };
    $substs = array_combine(array_map($keyCallback, array_keys($substs)), $substs);

    return nsUtils::Subst($aString, $substs);
  }

  /**********************************************************************************************************************
  * Universal Contains substring/array_key/array_value
  ***********************************************************************************************************************/
  public static function Contains(string|array $aHaystack, string|array $aNeedle, int $aPos = kZero) {
    $rv = false;
    if (is_string($aNeedle)) {
      $aNeedle = [$aNeedle];
    }

    foreach ($aNeedle as $_value) {
      if (is_array($aHaystack)) {
        $rv = ($aPos === 1) ? array_key_exists($_value, $aHaystack) : in_array($_value, $aHaystack);
      }
      else {
        switch ($aPos) {
          case nsUtils::STARTS_WITH:
            $rv = str_starts_with($aHaystack, $_value);
            break;
          case nsUtils::ENDS_WITH:
            $rv = str_ends_with($aHaystack, $_value);
            break;
          case nsUtils::CONTAINS:
          default:
            $rv = str_contains($aHaystack, $_value);
        }
      }

      if ($rv) {
        break;
      }
    }

    return $rv;
  }

  /**********************************************************************************************************************
  * Explode a string into an array
  ***********************************************************************************************************************/
  public static function ExplodeStr(string $aSeparator, string $aString) {
    return (!str_contains($aString, $aSeparator)) ? [$aString] :
            array_values(array_filter(explode($aSeparator, $aString), 'strlen'));
  }

  /**********************************************************************************************************************
  * Explode a path into an array
  ***********************************************************************************************************************/
  public static function ExplodePath(string $aPath) {
    return ($aPath == kSlash) ? ['root'] : nsUtils::ExplodeStr(kSlash, $aPath);
  }

  /**********************************************************************************************************************
  * Build a path from arguments
  ***********************************************************************************************************************/
  public static function BuildPath(...$aParts) {
    $parts = kEmptyArray;
    $path = strtr(implode(kSlash, $aParts), '\\', kSlash);
    $prefix = kEmptyString;
    $absolute = false;

    // extract a prefix being a protocol://, protocol:, protocol://drive: or simply drive:
    if (preg_match('{^( [0-9a-z]{2,}+: (?: // (?: [a-z]: )? )? | [a-z]: )}ix', $path, $match)) {
      $prefix = $match[1];
      $path = substr($path, strlen($prefix));
    }

    if (substr($path, 0, 1) === kSlash) {
      $absolute = true;
      $path = substr($path, 1);
    }

    $up = false;

    foreach (explode(kSlash, $path) as $chunk) {
      if (kDotDot === $chunk && ($absolute || $up)) {
        array_pop($parts);
        $up = !(empty($parts) || kDotDot === end($parts));
      }
      elseif (kDot !== $chunk && kEmptyString !== $chunk) {
        $parts[] = $chunk;
        $up = kDotDot !== $chunk;
      }
    }

    return $prefix . ($absolute ? kSlash : kEmptyString) . implode(kSlash, $parts);
  }

  /**********************************************************************************************************************
  * Strip the string of the supplied substring
  ***********************************************************************************************************************/
  public static function StripStr (string $aStr, string $aStrip = kEmptyString) {
    return str_replace($aStrip, kEmptyString, $aStr);
  }
  /**********************************************************************************************************************
  * Generate a random hexadecimal string
  *
  * @param $aLength   Desired number of final chars
  * @returns          Random hexadecimal string of desired length
  **********************************************************************************************************************/
  public static function HexString(int $aLength = 40) {
    return bin2hex(random_bytes(($aLength <= 1) ? 1 : (int)($aLength / 2)));
  }

  /********************************************************************************************************************
  * Hash a password
  *********************************************************************************************************************/
  public static function PasswordHash(string $aPassword, mixed $aCrypt = PASSWORD_BCRYPT, ?string $aSalt = null) {
    switch ($aCrypt) {
      case nsUtils::PASSWORD_CLEARTEXT:
        // We can "hash" a cleartext password by prefixing it with the fake algo prefix $clear$
        if (str_contains($aPassword, kDollar)) {
          // Since the dollar sign is used as an identifier and/or separator for hashes we can't use passwords
          // that contain said dollar sign.
          nsRuntime::Error('Cannot "hash" this Clear Text password because it contains a dollar sign.');
        }

        return kDollar . nsUtils::PASSWORD_CLEARTEXT . kDollar . time() . kDollar . $aPassword;
      case nsUtils::PASSWORD_HTACCESS:
        // We want to be able to generate Apache APR1-MD5 hashes for use in .htpasswd situations.
        $salt = $aSalt;

        if (!$salt) {
          $salt = kEmptyString;

          for ($i = 0; $i < 8; $i++) {
            $offset = hexdec(bin2hex(openssl_random_pseudo_bytes(1))) % 64;
            $salt .= nsUtils::APRMD5_ALPHABET[$offset];
          }
        }

        $salt = substr($salt, 0, 8);
        $max = strlen($aPassword);
        $context = $aPassword . kDollar . nsUtils::PASSWORD_HTACCESS . kDollar . $salt;
        $binary = pack('H32', md5($aPassword . $salt . $aPassword));

        for ($i = $max; $i > 0; $i -= 16) {
          $context .= substr($binary, 0, min(16, $i));
        }

        for ($i = $max; $i > 0; $i >>= 1) {
          $context .= ($i & 1) ? chr(0) : $aPassword[0];
        }

        $binary = pack('H32', md5($context));

        for ($i = 0; $i < 1000; $i++) {
          $new = ($i & 1) ? $aPassword : $binary;

          if ($i % 3) {
            $new .= $salt;
          }
          if ($i % 7) {
            $new .= $aPassword;
          }

          $new .= ($i & 1) ? $binary : $aPassword;
          $binary = pack('H32', md5($new));
        }

        $hash = kEmptyString;

        for ($i = 0; $i < 5; $i++) {
          $k = $i + 6;
          $j = $i + 12;
          if($j == 16) $j = 5;
          $hash = $binary[$i] . $binary[$k] . $binary[$j] . $hash;
        }

        $hash = chr(0) . chr(0) . $binary[11] . $hash;
        $hash = strtr(strrev(substr(base64_encode($hash), 2)), nsUtils::BASE64_ALPHABET, nsUtils::APRMD5_ALPHABET);

        return kDollar . nsUtils::PASSWORD_HTACCESS . kDollar . $salt . kDollar . $hash;
      default:
        // Else, our standard (and secure) default is PASSWORD_BCRYPT hashing.
        // We do not allow custom salts for anything using password_hash as PHP generates secure salts.
        // PHP Generated passwords are also self-verifiable via password_verify.
        return password_hash($aPassword, $aCrypt);
    }
  }

  /********************************************************************************************************************
  * Check a password
  *********************************************************************************************************************/
  public static function PasswordVerify(string $aPassword, string $aHash) {
    // We can accept a pseudo-hash for clear text passwords in the format of $clrtxt$unix-epoch$clear-text-password
    if (str_starts_with($aHash, kDollar . nsUtils::PASSWORD_CLEARTEXT)) {
      $password = nsUtils::ExplodeStr(kDollar, $aHash) ?? null;

      if ($password == null || count($password) > 3) {
        nsRuntime::Error('Unable to "verify" this Clear Text "hashed" password.');
      }

      return $aPassword === $password[2];
    }

    // We can also accept an Apache APR1-MD5 password that is commonly used in .htpasswd
    if (str_starts_with($aHash, kDollar . nsUtils::PASSWORD_HTACCESS)) {
      $salt = nsUtils::ExplodeStr(kDollar, $aHash)[1] ?? null;

      if(!$salt) {
        nsRuntime::Error('Unable to verify this Apache APR1-MD5 hashed password.');
      }

      return nsUtils::PasswordHash($aPassword, nsUtils::PASSWORD_HTACCESS, $salt) === $aHash;
    }

    // For everything else send to the native password_verify function.
    // It is almost certain to be a BCRYPT2 hash but hashed passwords generated BY PHP are self-verifiable.
    return password_verify($aPassword, $aHash);
  }

  /********************************************************************************************************************
  * Generates a v4 random guid or a "v4bis" guid with static vendor node
  *********************************************************************************************************************/
  public static function GlobalIdentifer(?string $aVendor = null, ?bool $aXPCOM = null) {
    if ($aVendor) {
      if (strlen($aVendor) < 3) {
        nsRuntime::Error('v4bis GUIDs require a defined vendor of more than three characters long.');
      }

      // Generate 8 pseudo-random bytes
      $bytes = random_bytes(8);

      // Knock the vendor down to lowercase so we can simply use a switch case
      $vendor = strtolower($aVendor);

      // We want "v4bis" GUIDs with the static vendor part to match the broken version of GUIDGenX for known nodes
      // as Moonchild half-assed his tool that he wrote for this and by the time it was discovered several were already
      // using the previous incorrect GUIDs.
      $knownVendorNodes = array(
        'binoc'           => hex2bin('8b97957ad5f8ea47'),
        'binoc-legacy'    => hex2bin('9aa0aa0e607640b9'),
        'mcp'             => hex2bin('bfc5fc555c87dbc4'),
        'lootyhoof'       => hex2bin('b98e98e62085837f'),
        'pseudo-static'   => hex2bin('93763763d1ad1978')
      );

      switch ($vendor) {
        case 'binoc':
        case 'binaryoutcast':
        case 'binary outcast':
          $bytes .= $knownVendorNodes['binoc'];
          break;
        case 'pseudo-static':
        case 'pseudostatic':
        case 'addons':
        case 'add-ons':
        case 'apmo':
          $bytes .= $knownVendorNodes['pseudo-static'];
          break;
        case 'mcp':
        case 'moonchildproductions':
        case 'moonchild productions':
          $bytes .= $knownVendorNodes['mcp'];
          break;
        case 'binoc-legacy':
        case 'lootyhoof':
          $bytes .= $knownVendorNodes[$vendor];
          break;
        default:
          // Since this isn't a known busted vendor node then we should generate it ourselves.
          // This matches the fixed version of GUIDGenX 1.1 which is to md5 hash the vendor string then
          // split it in half and XOR the two parts for the final value

          // XXXTobin: A future update could do a sha256 by halving it and again then XORing the two sets
          // and XORing the final two results.
          $vendor = hash('md5', $aVendor);
          $bytes .= hex2bin(substr($vendor, 0, 16)) ^ hex2bin(substr($vendor, 16, 32));
      }
    }
    else {
      // This is a pure v4 UUID spec which is 16 pseudo-random bytes.
      $bytes = random_bytes(16);
    }

    // Set the version and variant
    // NOTE: Like everything Moonzilla does, he did not set the variant value when he initially came up with "v4bis"
    // putting a busted generator into production use for the whole team. Sad!
    $bytes[6] = chr(ord($bytes[6]) & 0x0f | 0x40);
    $bytes[8] = chr(ord($bytes[8]) & 0x3f | 0x80);

    $hex = bin2hex($bytes);
    $guid = vsprintf('%s%s-%s-%s-%s-%s%s%s', str_split($hex, 4));

    // We want the GUID in XPIDL/C++ Header notation
    if ($aXPCOM) {
      $explode = nsUtils::ExplodeStr(kDash, $guid);
      $rv = "%{C++" . kNewLine . "//" . kSpace . kLeftBrace . $guid . kRightBrace . kNewLine .
            "#define NS_CHANGE_ME_IID" . kSpace . 
            vsprintf("{ 0x%s, 0x%s, 0x%s, { 0x%s, 0x%s, 0x%s, 0x%s, 0x%s, 0x%s, 0x%s, 0x%s } }",
                     [$explode[0], $explode[1], $explode[2],
                      substr($explode[3], 0, 2), substr($explode[3], 2, 2),
                      substr($explode[4], 0, 2), substr($explode[4], 2, 2), substr($explode[4], 4, 2),
                      substr($explode[4], 6, 2), substr($explode[4], 8, 2), substr($explode[4], 10, 2)]) . kNewLine .
            "%}";
    }
    else {
      // We like Microsoft GUID notation not UUID which means Lisa needs braces.. I mean the GUID.
      $rv = '{' . $guid . '}';
    }

    return $rv;
  }
}

// == | Static Output Class | =========================================================================================

class nsConsole {
  const PHP_ERROR_CODES = array(
    E_ERROR                   => 'System Error',
    E_WARNING                 => 'System Warning',
    E_PARSE                   => 'System Error (Parser)',
    E_NOTICE                  => 'System Notice',
    E_CORE_ERROR              => 'System Error (Core)',
    E_CORE_WARNING            => 'System Warning (Core)',
    E_COMPILE_ERROR           => 'System Error (Compiler)',
    E_COMPILE_WARNING         => 'System Warning (Compiler)',
    E_USER_ERROR              => 'System Error (Application)',
    E_USER_WARNING            => 'System Warning (Application)',
    E_USER_NOTICE             => 'System Notice (Application)',
    E_STRICT                  => 'System Error (Strict)',
    E_RECOVERABLE_ERROR       => 'System Error (Recoverable)',
    E_DEPRECATED              => 'System Deprecation',
    E_USER_DEPRECATED         => 'System Deprecation (Application)',
    E_ALL                     => 'Unable to Comply',
    E_EXCEPTION               => 'System Exception',
  );

  const HTTP_HEADERS = array(
    404                       => 'HTTP/1.1 404 Not Found',
    501                       => 'HTTP/1.1 501 Not Implemented',
  );

  const HTTP_STATUS_CODE = array(
    200       => 'OK',
    204       => 'No Content',
    400       => 'Bad Request',
    401       => 'Unauthorized',
    402       => 'Payment Required',
    403       => 'Forbidden',
    404       => 'Not Found',
    405       => 'Method Not Allowed',
    406       => 'Not Acceptable',
    415       => 'Unsupported Media Type',
    418       => 'I\'m a teapot',
    500       => 'Internal Server Error',
    501       => 'Not Implemented',
    502       => 'Bad Gateway',
    503       => 'Service Unavailable',
  );

  const MIME_TYPES = array(
    'text'    => 'text/plain',
    'html'    => 'text/html',
    'rdf'     => 'text/rdf',
    'xul'     => 'application/vnd.mozilla.xul+xml',
    'xhtml'   => 'application/xhtml+xml',
    'xml'     => 'text/xml',
    'css'     => 'text/css',
    'js'      => 'application/js',
    'json'    => 'application/json',
    'bin'     => 'application/octet-stream',
    'xpi'     => 'application/x-xpinstall',
    'jar'     => 'application/java-archive',
    '7z'      => 'application/x-7z-compressed',
    'xz'      => 'application/x-xz',
    'zip'     => 'application/zip',
  );

  /**********************************************************************************************************************
  * Error handling
  **********************************************************************************************************************/
  public static function EnableErrorHandler() {
    set_error_handler(__CLASS__ . kScopeOperator . "phpErrorHandler");
    set_exception_handler(__CLASS__ . kScopeOperator . "phpExceptionHandler");
  }

  public static function DisableErrorHandler() { restore_error_handler(); restore_exception_handler(); }

  public static function phpErrorHandler($eCode, $eMessage, $eFile, $eLine) {
    if (!(error_reporting() & $eCode)) {
      // Don't do jack shit because the developers of PHP think users shouldn't be trusted.
      return;
    }

    self::Report(['code' => $eCode, 'message' => $eMessage,
                  'file' => $eFile, 'line' => $eLine,
                  'trace' => debug_backtrace(2)]);
  }

  public static function phpExceptionHandler($ex) {
    self::Report(['code' => E_EXCEPTION, 'message' => $ex->getMessage(),
                  'file' => $ex->getFile(), 'line' => $ex->getLine(),
                  'trace' => $ex->getTrace()]);
  }

  /**********************************************************************************************************************
  * Sends HTTP Headers to client using a short name
  *
  * @dep HTTP_HEADERS
  * @dep kDebugMode
  * @dep nsRuntime::Error()
  * @param $aHeader    Short name of header
  **********************************************************************************************************************/
  public static function Header(string|int $aHeader, bool $aSendAllOnContentType = false) { 
    if (is_int($aHeader) && nsUtils::Contains(self::HTTP_STATUS_CODE, $aHeader, 1)) {
      self::HttpStatusCode($aHeader);
      
      if ($aHeader < 400) {
        return true;
      }
      else {
        self::SendHeaders();
        exit();
      }
    }

    if (nsUtils::Contains(self::MIME_TYPES, $aHeader, 1)) {
      self::ContentType($aHeader);

      if ($aSendAllOnContentType) {
        return self::SendHeaders();
      }
      else {
        return true;
      }
    }

    return nsRuntime::SetConfig('console.output.headers[]', $aHeader);
  }

  /********************************************************************************************************************
  * Send HTTP Headers
  ********************************************************************************************************************/
  public static function SendHeaders() {
    $responseCode = nsRuntime::GetConfig('console.output.responseCode', 200);
    nsRuntime::SetConfig('console.output.headers[]', 'Content-type'. kColon . kSpace . nsRuntime::GetConfig('console.output.contentType'));
    nsRuntime::SetConfig('console.output.headers[]', 'HTTP/1.1' . kSpace . $responseCode . kSpace . self::HTTP_STATUS_CODE[$responseCode]);

    $headers = nsRuntime::GetConfig('console.output.headers', kEmptyArray);

    foreach ($headers as $_value) {
      header(trim($_value), true);
    }
  }

  /********************************************************************************************************************
  * Gets or sets the status code
  ********************************************************************************************************************/
  public static function HttpStatusCode(?string $aStatusCode = null) {
    if (!$aStatusCode) {
      nsRuntime::GetConfig('console.output.responseCode', 200);
    }

    if (nsUtils::Contains(self::HTTP_STATUS_CODE, $aStatusCode, 1)) {
      nsRuntime::SetConfig('console.output.responseCode', $aStatusCode);
    }
  }

  /********************************************************************************************************************
  * Gets or sets the "default" content type so we don't have to output the header ourselves in most cases.
  ********************************************************************************************************************/
  public static function ContentType(?string $aContentType = null) {
     if ($aContentType === null) {
      return nsRuntime::GetConfig('console.output.contentType');
    }

    if (nsUtils::Contains(self::MIME_TYPES, $aContentType, 1)) {
      return nsRuntime::SetConfig('console.output.contentType', self::MIME_TYPES[$aContentType]);
    }
  }

  /**********************************************************************************************************************
  * Sends HTTP Header to redirect the client to another URL
  *
  * @param $aURL   URL to redirect to
  **********************************************************************************************************************/
  public static function Redirect($aURL) { header('Location: ' . $aURL, true, 302); exit(); }

  /**********************************************************************************************************************
  * Get a subdomain or base domain from a host
  *
  * @dep kDot
  * @dep nsUtils::ExplodeStr()
  * @param $aHost       Hostname
  * @param $aReturnSub  Should return subdmain
  * @returns            domain or subdomain
  ***********************************************************************************************************************/
  public static function GetDomain(string $aHost, ?bool $aReturnSub = null) {
    $host = nsUtils::ExplodeStr(kDot, $aHost);
    return implode(kDot, $aReturnSub ? array_slice($host, 0, -2) : array_slice($host, -2, 2));
  }

  /**********************************************************************************************************************
  * var_export() + exit()
  **********************************************************************************************************************/
  public static function Export($aVariable) { print(var_export($aVariable, true)); exit(); }

  /******************************************************************************************************************
  * Output details about a failure condition
  ******************************************************************************************************************/
  public static function Report(array $aMetadata) {
    if (nsRuntime::Debug() && nsRuntime::SuperGlobal('get', 'runtime')) {
      nsConsole::Output((nsConsole::PHP_ERROR_CODES[$aMetadata['code']] ?? nsConsole::PHP_ERROR_CODES[E_ALL]) . kSpaceDashSpace . $aMetadata['message']);
    }

    $traceline = fn($eFile, $eLine) => str_replace(kRootPath, kEmptyString, $eFile) . kColon . $eLine;
    $functions = ['phpErrorHandler', 'phpExceptionHandler', 'trigger_error'];
    $trace = ($aMetadata['file'] && $aMetadata['line']) ? [$traceline($aMetadata['file'], $aMetadata['line'])] : kEmptyArray;

    foreach ($aMetadata['trace'] as $_key => $_value) {
      if (in_array($_value['function'], $functions)) {
        continue;
      }

      $trace[] = $traceline($_value['file'], $_value['line']);
    }

    $title = nsConsole::PHP_ERROR_CODES[$aMetadata['code']] ?? nsConsole::PHP_ERROR_CODES[E_ALL];
    $content = $aMetadata['message'];

    if (!SAPI_IS_CLI) {
      $content = '<h2>Traceback</h2><ul>';

      foreach ($trace as $_value) {
        $content .= '<li>' .  str_replace(kSlash, kDot, str_replace('.php', kEmptyString, ltrim($_value, kSlash))) . '</li>';
      }

      $content .= '</ul>';
      $content .= is_string($aMetadata['message']) ?
                 '<hr /><p><strong>' . $aMetadata['message'] . '</strong></p>':
                 kEmptyString;

      $commandBar = ['onclick="history.back()"' => 'Go Back'];

      if (nsRuntime::Component(kSpecialComponent) || !nsRuntime::GetConfig('constant.components.site')) {
        nsRuntime::SetConfig('console.ui.commandbar', array_merge($commandBar, ['/special/' => kSpecialComponentName]));
      }
      else {
        nsRuntime::SetConfig('console.ui.commandbar', array_merge($commandBar, kDefaultMenu));
      }

      nsRuntime::SetConfig('console.ui.sectionName', kEmptyString);
      nsConsole::Content($content, ['title' => $title, 'statustext' => '<span class="pulseText">Please contact a system administrator.</span>']);
    }

    nsConsole::Output(['title'=> $title, 'content' => ['errorMessage' => $content, 'traceback' => $trace]]);
  }

  /******************************************************************************************************************
  * Simply prints output and sends header if not cli and exits
  ******************************************************************************************************************/
  public static function Output($aContent, ?string $aHeader = 'text') {
    $store = nsRuntime::GetStore();
    $content = null;

    if ($aContent === true) {
      $content = nsRuntime::EnsureValue($store['console']['output']['responseBody']);
    }

    if (nsRuntime::Debug() && nsRuntime::SuperGlobal('get', 'runtime')) {
      $finalContent = $store;
      $finalContent['console']['output']['responseBody'] = $content ?? $aContent;
      $finalContent['console']['output']['responseTime'] = microtime(true) - nsRuntime::SuperGlobal('server', 'REQUEST_TIME_FLOAT', 0);
      $finalContent = json_encode($finalContent, nsUtils::JSON_ENCODE_FLAGS['display']);
      self::Header('text', true);
      print($finalContent);
      exit();
    }

    if (is_array($aContent)) {
      $title = $aContent['title'] ?? 'Output';
      $content = $aContent['content'] ?? kEmptyString;

      if ($title == 'Output' && $content == kEmptyString) {
        $content = $aContent ?? $content;
      }
    }
    else {
      $title = 'Output';
      $content = $content ?? $aContent ?? kEmptyString;
    }

    $content = (is_string($content) || is_int($content)) ? $content : json_encode($content, nsUtils::JSON_ENCODE_FLAGS['display']);

    // Send the header if not cli
    if (SAPI_IS_CLI) {
      if (!CLI_NO_LOGO) {
        $software = $title . kSpaceDashSpace . kAppVendor . kSpace . kAppName . kSpace . kAppVersion;
        $titleLength = 120 - 8 - strlen($software);
        $titleLength = ($titleLength > 0) ? $titleLength : 2;
        $title = kNewLine . '==' . kSpace . kPipe . kSpace . $software . kSpace . kPipe . kSpace . str_repeat('=', $titleLength);
        $content = $title . kNewLine . kNewLine . $content . kNewLine . kNewLine . str_repeat('=', 120) . kNewLine;
      }
    }
    else {
      self::ContentType($aHeader);
      self::SendHeaders();
    }

    // Write out the content
    print($content);

    // We're done here...
    exit();
  }

  /******************************************************************************************************************
  * Basic Site Content Generation using a Special Template
  ******************************************************************************************************************/
  public static function Content(mixed $aContent, array $aMetadata = kEmptyArray) {
    $template = nsRuntime::ReadFile(kDefaultSkinPath . kSlash . 'template.xhtml');
    $stylesheet = nsRuntime::ReadFile(kDefaultSkinPath . kSlash . 'stylesheet.css');

    if (!$template || SAPI_IS_CLI) {
      nsConsole::Output(['content' => $aContent, 'title' => $aMetadata['title'] ?? 'Output']);
    }

    $content = $aContent;

    $metadata = function($val) use(&$aMetadata) {
      return $aMetadata[$val] ?? null;
    };

    $menuize = function($aMenu) {
      $rv = kEmptyString;

      foreach ($aMenu as $_key => $_value) {
        if (nsUtils::Contains($_key, 'onclick=', 1)) {
          $rv .= '<li><a href="#"' . kSpace . $_key . '>' . $_value . '</a></li>';
        }
        else {
          $rv .= '<li><a href="' . $_key . '">' . $_value . '</a></li>';
        }
      }

      return $rv;
    };

    if ((is_string($content) || is_int($content)) && !$metadata('textbox') && !$metadata('iframe')) {
      if (!nsUtils::Contains($content, ['<p', '<ul', '<ol', '<h1', '<h2', '<h3', '<table'])) {
        $content = '<p>' . $content . '</p>';
      }
    }
    else {
      $aMetadata['textbox'] = true;
    }

    if ($metadata('textbox')) {
      $content = (is_string($content) || is_int($content)) ? $content : json_encode($content, nsUtils::JSON_ENCODE_FLAGS['display']);
      $content = '<form><textarea class="special-textbox" name="content" rows="30" readonly>' . $content . '</textarea></form>';
    }

    $siteName = nsRuntime::GetConfig('console.ui.siteName', kAppName);
    $sectionName = nsRuntime::GetConfig('console.ui.sectionName', kEmptyString);

    if ($sectionName) {
      $siteName = $sectionName . kSpaceDashSpace . $siteName;
    }

    $isTestCase = (!$metadata('title') && nsRuntime::GetConfig('special.testCase') && nsRuntime::Component(kSpecialComponent));

    $substs = array(
      '{$SITE_STYLESHEET}'  => $stylesheet ?? kEmptyString,
      '{$PAGE_CONTENT}'     => $content,
      '{$SITE_DOMAIN}'      => nsRuntime::SuperGlobal('server', 'SERVER_NAME') ?? 'localhost',
      '{$SITE_NAME}'        => $siteName,
      '{$SITE_MENU}'        => $menuize(nsRuntime::GetConfig('console.ui.commandbar')),
      '{$SITE_SECTION}'     => $sectionName ?? kEmptyString,
      '{$PAGE_TITLE}'       => $isTestCase ? '[Test]' . kSpace . nsRuntime::GetConfig('special.testCase') : ($metadata('title') ?? 'Output'),
      '{$PAGE_STATUS}'      => $metadata('statustext') ?? nsRuntime::GetConfig('console.ui.statustext'),
      '{$SKIN_PATH}'        => substr(kDefaultSkinPath, 1),
      '{$SOFTWARE_VENDOR}'  => kAppVendor,
      '{$SOFTWARE_NAME}'    => kAppName,
      '{$SOFTWARE_VERSION}' => kAppVersion,
    );

    $content = nsUtils::Subst($template, $substs);

    @ob_end_clean();
    nsConsole::Output($content, 'html');
  }

  /**********************************************************************************************************************
  * Create an XML Document 
  ***********************************************************************************************************************/
  public static function extensibleMarkup(string|array $aData, ?bool $aDirectOutput = null) {
    if (is_string($aData)) {
      $xml = $aData;
    }
    else {
      $dom = new DOMDocument('1.0');
      $dom->encoding = "UTF-8";
      $dom->formatOutput = true;

      $processChildren = function($aData) use (&$dom, &$processChildren) {
        if (!($aData['@elem'] ?? null)) {
          return false;
        }

        // Create the element
        $element = $dom->createElement($aData['@elem']);

        // Properly handle content using XML and not try and be lazy.. It almost never works!
        if (array_key_exists('@content', $aData) && is_string($aData['@content'])) {
          if (nsUtils::Contains($aData['@content'], ["<", ">", "?", "&", "'", "\""])) {
            $content = $dom->createCDATASection($aData['@content'] ?? kEmptyString);
          }
          else {
            $content = $dom->createTextNode($aData['@content'] ?? kEmptyString);
          }

          $element->appendChild($content);
        }
       
        // Add any attributes
        if (!empty($aData['@attr']) && is_array($aData['@attr'])) {
          foreach ($aData['@attr'] as $_key => $_value) {
            $element->setAttribute($_key, $_value);
          }
        }
       
        // Any other items in the data array should be child elements
        foreach ($aData as $_key => $_value) {
          if (!is_numeric($_key)) {
            continue;
          }
       
          $child = $processChildren($_value);
          if ($child) {
            $element->appendChild($child);
          }
        }
       
        return $element;
      };

      $child = $processChildren($aData);
      $xml = null;

      if ($child) {
        $dom->appendChild($child);
      }

      $xml = $dom->saveXML();

      if (!$xml) {
        nsRuntime::Error('Could not generate extensible markup.');
      }
    }

    if ($aDirectOutput) {
      nsConsole::Output($xml, 'xml');
    }

    return $xml;
  }
}

// ====================================================================================================================

// An application has been detected - Unimatrix 424, Grid 116 - activate.
nsRuntime::init();

// ====================================================================================================================